%Event-Trggered control of a Linear system with bounded data rate
%non-instantaneous communication with zero disturbance
%time varying (piece-wise constant communication rate)
%intermittent channel blackouts

%conference version

%Author: Pavan Tallapragada

%function ET_no_blackout()
function [ TH,X,TJ,XJ,OP ] = ET_blackout_conf(pmax, varargin)

if nargin >= 2
    save_plot = 1;
else
    save_plot = 0;
end

%%%%%%%%% Params %%%%%%%%%%%%
%%

%{
A = [[1.38, -0.2077 6.715 -5.676]; ...
     [-0.5814 -4.29 0 0.675]; ...
     [1.067 4.273 -6.654 5.893]; ...
     [0.048 4.273 1.343 -2.104] ];
B = [ 0 0; 5.679 0; 1.136 -3.146; 1.136 0];
K = -place(A,B,[-3 -3 -2 -2])
%}
%
%%
A = [2 0; 1 1];
B = [1; 0];
K = -place(A,B,[-1 -2]);
%

[n,n] = size(A);
[n,m] = size(B);

Q = eye(n);

Abar = A+B*K;
P = lyap(Abar', Q);

Lm_Q = min(eig(Q));
LM_P = max(eig(P));
Lm_P = min(eig(P));

beta = 0.8*(Lm_Q/LM_P);
a = 1.2;
sigma = 0.9;
sigma1 = 0.8;

T0 = 0;
%%
%{
Tslot = 1e-2*(1/gain)*[ 2, 3, 2, 3, 3, 1, 2, 3, 3 ];
Islot = cumsum(Tslot) + T0;
Pbar = floor( [ 1, 0.5, 0.2, 0, 0.4, 0.1, 0.2, 0, 0.2 ]*pmax );
Rratio = 1 + 1*[ 0.8, 0.1, 0.2, 0.0001, 0.4, 0.1, 0.01, 0.0001, 0.1 ];

Nslots = length(Pbar);
%}

%gen_channel
load channels/channel_var1

nbs = inf*ones(size(Pbar));     %next blackout slot

slotnum = 1:Nslots;
ind_black = slotnum(Pbar==0);
for k = 1:Nslots    
    nbstmp = min( ind_black(ind_black > k) );
    if isempty(nbstmp)    %no blackout after slot k
        nbs(k) = inf;
    else
        nbs(k) = nbstmp;
    end
end

Tf = 0.9*Islot(end)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%initial conditions
%xp0 = [5, -3, -4, 4]';
%xh0 = [4.9, -2.9, -4.1, 4.1]';

xp0 = [6, -4]';
xh0 = [0, 0]';

de0 = 1.2*norm(xp0 - xh0, Inf);
Vd0 = 1.3*xp0'*P*xp0;
p_Lbnd0 = 0;
p0 = 0;
k0 = 0;
t_jmp0 = 0;
r_jmp0 = 0;
rtil_jmp0 = 0;
z0 = zeros(n, 1);
zD0 = zeros(n, 1);
del0 = de0;
intrans0 = 0;    %0: no message in transit, 1: message in transit
                 %2: message received, but waiting to update control
NUT0 = 0;        %next control update time to determine rtil_jmp
t_nxt_psi0 = 0;  %time of next jump in psitau. this is only place holder,
                 %will be intialized later

TH = [ ];
X = [ ];
TJ = [ ];
XJ = [ ];
OP = [ ];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

w = (Lm_Q/LM_P) - beta;
W = (Lm_Q/LM_P) - a*beta;
c = (W*sqrt(Lm_P))/(2*sqrt(n)*norm(P*B*K));

if W <= 0
    disp('check W');
    return;
end

alpha_A = norm(A);
theta = alpha_A + beta/2;
theta_bar = norm(A, inf) + beta/2;

%c1 = ( 2*norm(P)*nu ) / ( sqrt(Lm_P*V0) );
%c2 = ( W*nu ) / ( c*alpha_A*sqrt(V0) );
c1 = 0;
c2 = 0;

%---------------------------

Params.save_plot = save_plot;

Params.T0 = T0;
Params.Tf = Tf;
Params.n = n;
Params.m = m;
Params.A = A;
Params.B = B;
Params.K = K;
Params.P = P;
Params.Q = Q;
Params.beta = beta;
Params.w = w;
Params.W = W;
Params.theta = theta;
Params.theta_bar = theta_bar;
Params.c = c;
Params.sigma1 = sigma1;

Params.Islot = Islot;
Params.Tslot = Tslot;
Params.Pbar = Pbar;
Params.nbs = nbs;

%---------------------------

G1 = Gamma1(1, 1, Tf, Params)
if G1 < 0
    disp('check G1')
    return;
end

T = 0.1*G1 %The parameter in the rho function

%coefficient of (1 - h0) in the definition of the rho function
c3 = ( w + theta )/( W * ( exp( (w + theta)*T ) - 1 ) );

%---------------------------

Params.c3 = c3;
Params.G1 = G1;
Params.T = T;

%---------------------------

TM = zeros(1, pmax);
for k = 1:pmax
    G2_unif_k = Gamma2_uniform( k, Params );
    TM(k) = 0.04*min([ G1, T, G2_unif_k ]);
end
%TM(k+1:end) = TM(k);
Rmin = max( [3:pmax+2]./TM )    % +2 in the numerator according to assump

%min R to ensure J = 0
Rmin_J0 = [ Pbar(1:end-1)./Tslot(2:end), Rmin ]

R = Rratio.*max(Rmin_J0, Rmin)

%del_t = 0.95*min([Tslot, 1/max(R(Pbar>0))])
%---------------------------

Params.TM = TM;
Params.R = R;
Params.Pbar = Pbar;
%Params.del_t = del_t;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

V_t0 = xp0'*P*xp0;
b0 = V_t0/Vd0;
eps0 = de0/(c*sqrt(Vd0));
psi0 = eps0 / (2^Pbar(1));
hch0 = eps0/rho(b0, c3)

[ isflag, L3tu, Pc_nb, DCtogo, eps_r ] = ...
           is_channel_control_comp_J0( eps0, Params );

L3tu
eps_r = eps_r
if isflag == 0
    disp('check params, L3');
    disp(L3tu )
    return;
end

%---------------------------

%Params.phiN = Pc_nb;
Params.Pc_nb = Pc_nb;
Params.DCtogo = DCtogo;
Params.eps_r = eps_r;

%---------------------------

slt = slot_at(T0, Params);
psitau_0 = psitaufun(T0, Params);

hpf_init_val = b0
btil_init_val = b_tilde(TM(Pbar(1)), b0, eps0, Params)

if btil_init_val < 0 || btil_init_val > 1
    disp('check params, b_tilde(TM, b0, eps0, Params)');
    disp( b_tilde(TM(Pbar(1)), b0, eps0, Params) )
    return;
end

hch_init_val = hch_bar(0, b0, eps0, 0, Params)
Lc2_init_val = Lctil2_fun(T0, b0, eps0, Params)

if Lc2_init_val < 0 || Lc2_init_val > 1
    disp('check params, Lc2_fun(T0, T0, b0, eps0, Params)');
    disp( Lctil2_fun(T0, T0, b0, eps0, Params) )    
    return;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%

t_nxt_psi0 = time_next_psitau_jmp(T0, Params);

%aggregate state of the hybrid system
x0 = [xp0; xh0; Vd0; de0; p_Lbnd0; p0; k0; t_jmp0; r_jmp0; rtil_jmp0; ...
      z0; zD0; del0; intrans0; NUT0; t_nxt_psi0];

opt = odeset('reltol',1e-12,'abstol',1e-12,...%'maxstep',0.9*delta_t,...
             'events',@comm_event);

TermId = [ ];

q = 3;
Pat = cell(q,1);

Pat{1} = 1:4*n+12;
Pat{2} = 1:4*n+12;
Pat{3} = 1:4*n+12;
Pat{4} = 1:4*n+12;

[ TH,X,TJ,XJ,OP ]=hybrid_sys( @OL_control, ...
                              @fin_dat_comm_jmp, @comm_event, ...
                              @outputs_fun, TermId, Pat, T0, Tf, x0, ...
                              opt, Params );

to_plot = 1;
if to_plot == 1
    plot_fun( TH, X, TJ, XJ, OP, Params );
end

%-------------------------------------------------------

function tslot = slot_at(t, Par)

Islot = Par.Islot;

numslot = 1:length(Islot);

tmp = numslot( (t - Islot) <= 0 );
tslot = tmp(1);

%-------------------------------------------------------

function [ psitau, Phi ] = psitaufun( t, Par )

Islot = Par.Islot;
Tslot = Par.Tslot;
Pbar = Par.Pbar;
R = Par.R;
Pc_nb = Par.Pc_nb;

slt_t = slot_at(t,Par);

Phi = max(0, floor( Pc_nb(slt_t) - ...
      R(slt_t)*(t - (Islot(slt_t)-Tslot(slt_t)) ) ) );
psitau = min( Pbar(slt_t), Phi );

%-------------------------------------------------------

function t_nxt_psi = time_next_psitau_jmp(t, Par)

Islot = Par.Islot;
Tslot = Par.Tslot;
Pbar = Par.Pbar;
R = Par.R;
Pc_nb = Par.Pc_nb;

slt = slot_at(t, Par);
pbar = Pbar(slt);

t_nxt_slot = Islot(slt);

[ psitau, Phi ] = psitaufun( t, Par );

if psitau == 0
    t_nxt_psi = t_nxt_slot;
else
    t_nxt_Phi = (Pc_nb(slt)-psitau+1)/R(slt) + Islot(slt) - Tslot(slt);
    t_nxt_psi = min(t_nxt_Phi, t_nxt_slot);
end

%-------------------------------------------------------

function btild_val = b_tilde(tau, b0, eps0, Par)

w = Par.w;
W = Par.W;
theta = Par.theta;

f1 = b0 + W*eps0*( exp((w+theta)*tau) - 1 )/(w+theta);
f2 = exp(w*tau);

btild_val = f1/f2;

%-------------------------------------------------------

function hch_bar_val = hch_bar(tau, b0, eps0, p0, Par)

A = Par.A;
beta = Par.beta;
c3 = Par.c3;

b = b_tilde(tau, b0, eps0, Par);

num = norm(expm(A*tau), Inf) * exp(beta*tau/2) * eps0 ;

hch_bar_val = num /( (2^p0) * rho(b, c3) );

%-------------------------------------------------------

function Lctil1_val = Lctil1_fun(t, b0, eps0, Par)

TM = Par.TM;
R = Par.R;

psitau_t = psitaufun(t, Par);

if psitau_t == 0
    Tc = 2/R(slot_at(t,Par));
else
    Tc = TM(psitau_t);
end

Lctil1_val = b_tilde(Tc, b0, eps0, Par);

%-------------------------------------------------------

function Lctil2_val = Lctil2_fun(t, b0, eps0, Par)

TM = Par.TM;
R = Par.R;

psitau_t = psitaufun(t, Par);

if psitau_t == 0
    Tc = 2/R(slot_at(t,Par));
else
    Tc = TM(psitau_t);
end

Lctil2_val = hch_bar(Tc, b0, eps0, psitau_t, Par);

%-------------------------------------------------------

function dx = OL_control(t,x,Par)

n = Par.n;
A = Par.A;
B = Par.B;
K = Par.K;
beta = Par.beta;

xp = x(1:n);
xh = x(n+1:2*n);
Vd = x(2*n+1);
de = x(2*n+2);
p_Lbnd = x(2*n+3);
p = x(2*n+4);
k = x(2*n+5);
t_jmp = x(2*n+6);   %last transmit time
r_jmp = x(2*n+7);   %last reception time
rtil_jmp = x(2*n+8);    %last update time
z = x(2*n+8+1:3*n+8);
zD = x(3*n+8+1:4*n+8);
del = x(4*n+9);
intrans = x(4*n+10);
NUT = x(4*n+11);
t_nxt_psi = x(4*n+12);

dxp = A*xp + B*K*xh;
dxh = (A+B*K)*xh;
dVd = -beta*Vd;
dde = 0;
dp_Lbnd = 0;
dp = 0;
dk = 0;
dt_jmp = 0;
dr_jmp = 0;
drtil_jmp = 0;

dz = zeros(n,1);
dzD = zeros(n,1);
ddel = 0;
dintrans = 0;
dNUT = 0;
dt_nxt_psi = 0;

dx = [dxp; dxh; dVd; dde; dp_Lbnd; dp; dk; dt_jmp; dr_jmp; drtil_jmp; ...
      dz; dzD; ddel; dintrans; dNUT; dt_nxt_psi];

%-------------------------------------------------------
function x_new = fin_dat_comm_jmp(th, x, Par, jump_flag)

n = Par.n;
A = Par.A;
B = Par.B;
K = Par.K;
c = Par.c;
c3 = Par.c3;
Pbar = Par.Pbar;
P = Par.P;
TM = Par.TM;

Islot = Par.Islot;
Pc_nb = Par.Pc_nb;

Abar = A + B*K;

xp = x(1:n);
xh = x(n+1:2*n);
Vd = x(2*n+1);
de = x(2*n+2);
p_Lbnd = x(2*n+3);
p = x(2*n+4);
k = x(2*n+5);
t_jmp = x(2*n+6);   %last transmit time
r_jmp = x(2*n+7);   %last reception time
rtil_jmp = x(2*n+8);    %last update time
z = x(2*n+8+1:3*n+8);
zD = x(3*n+8+1:4*n+8);
del = x(4*n+9);
intrans = x(4*n+10);
NUT = x(4*n+11);
t_nxt_psi = x(4*n+12);

if jump_flag(1) >= 0   %means transmit jump
    
    k = k + 1;
    
    %de just before jumping
    de_ = norm(expm(A*(th(1)-t_jmp)),Inf)*del;

    V = xp'*P*xp;
    b = V/Vd;

    eps = de_ / ( c*sqrt(Vd) );
    
    slt = slot_at(th(1), Par);       %slot at current time
    psitau = psitaufun( th(1), Par );
       
    for jj = 1:psitau
        %psi = eps/(2^jj);
        hch_bar_val = hch_bar(TM(jj), b, eps, jj, Par);
        if hch_bar_val <= 0 
            disp('hch_bar negative where it should not be');            
            1;
        end
        if hch_bar_val <= 1
            break;
        end
    end
    %
    p_Lbnd = psitau;
    hch_bar_val = hch_bar(TM(p_Lbnd), b, eps, p_Lbnd, Par);
    if hch_bar_val <= 0
        disp('hch_bar negative where it should not be');        
        1;
    end
    %
    
    %p_Lbnd = jj;
    p = p_Lbnd;
        
    del = de_;
    
    z = xh;
    zD = xh;

    for i = 1 : p
        del = del / 2;
        zD = zD - del*((xp-zD)<=0) + del*((xp-zD)>0);
    end
    
    t_jmp = th(1);
    intrans = 1;    %message in transit
end

if jump_flag(2) >= 0   %means reception jump
    
    %compute the next control update time
    slt = slot_at(th(1), Par);       %slot at current time
    psitau = psitaufun( th(1), Par );
    
    if psitau >= 1 || Pbar(slt) == 0
        %can update right now        
        NUT = th(1);
    else
        %update at the beginning of the next slot
        NUT = Islot(slt);
    end
    
    r_jmp = th(1);
    intrans = 2;
end        

if jump_flag(3) >= 0   %means control update jump
    Deltatil = th(1) - t_jmp;
    xh = expm(Abar*Deltatil)*z + expm(A*Deltatil)*(zD - z);
    
    de = norm(expm(A*(Deltatil)),Inf)*del;
    
    V = xp'*P*xp;
    b = V/Vd;
    eps = de / ( c*sqrt(Vd) );
    if eps/rho(b, c3) > 1
        disp('hch(rtil_k) > 1');
        1;
    end
    
    rtil_jmp = th(1);
    intrans = 0;    %all messages recieved
end

if jump_flag(4) >= 0
    t_del = 1e-8;
    t_nxt_psi = time_next_psitau_jmp(th(1)+t_del, Par);
end

x_new = [xp; xh; Vd; de; p_Lbnd; p; k; t_jmp; r_jmp; rtil_jmp; z; zD; ...
         del; intrans; NUT; t_nxt_psi];

%---------------------------------------------------------
function [value,isterminal,direction] = comm_event(t,x,Par)

t_eps = 1e-8;
tp = t + t_eps;

n = Par.n;
A = Par.A;
P = Par.P;
c = Par.c;

TM = Par.TM;

R = Par.R;

Pc_nb = Par.Pc_nb;
DCtogo = Par.DCtogo;
eps_r = Par.eps_r;

xp = x(1:n);
xh = x(n+1:2*n);
Vd = x(2*n+1);
de = x(2*n+2);
p_Lbnd = x(2*n+3);
p = x(2*n+4);
k = x(2*n+5);
t_jmp = x(2*n+6);   %last transmit time
r_jmp = x(2*n+7);   %last reception time
rtil_jmp = x(2*n+8);    %last update time
z = x(2*n+8+1:3*n+8);
zD = x(3*n+8+1:4*n+8);
del = x(4*n+9);
intrans = x(4*n+10);
NUT = x(4*n+11);
t_nxt_psi = x(4*n+12);

xe = xp - xh;
V = xp'*P*xp;

b = V/Vd;
de_cur = norm(expm(A*(t-t_jmp)),Inf)*del;
eps = de_cur / (c*sqrt(Vd));

slt = slot_at(t, Par);
[ psitau_t, Phi_t ] = psitaufun( t, Par );
dctg_t = DCtogo(slt) - n*Pc_nb(slt) + n*Phi_t;

sltp = slot_at(tp, Par);
[ psitau_tp, Phi_tp ] = psitaufun( tp, Par );
dctg_tp = DCtogo(sltp) - n*Pc_nb(sltp) + n*Phi_tp;

term = -1*ones(6, 1);

term(1) = Lctil1_fun(t, b, eps, Par) - 1;
term(2) = Lctil1_fun(tp, b, eps, Par) - 1;
term(3) = Lctil2_fun(t, b, eps, Par) - 1;
term(4) = Lctil2_fun(tp, b, eps, Par) - 1;
term(5) = L3fun( t, eps, slt, eps_r(slt), dctg_t, Par ) - 1;
term(6) = L3fun( tp, eps, sltp, eps_r(sltp), dctg_tp, Par ) - 1;

if psitau_t == 0
    mtm = -1;
else
    mtm = max(term(term~=inf));
end

%
if t >= 0 && intrans == 0
    nxe = norm(xe,inf);
    if nxe > de_cur
        disp('de does not bound the encoding error')
        1;
    end
end
%

if intrans == 0    
    value(1) = mtm;
    value(2) = - 1;
    value(3) = - 1;
elseif intrans == 1
    
     if b >= 1
         1;
     end
    
    sltjmp = slot_at(t_jmp, Par);
  
    value(1) = -1;
    value(2) = (t - t_jmp) - p/R(sltjmp);
    value(3) = (t - t_jmp) - (p+2)/R(sltjmp);
elseif intrans == 2   
    value(1) = -1;
    value(2) = -1;
    value(3) = t - NUT;
end

value(4) = t - (t_nxt_psi - 1e-10);

value = value(:);

isterminal = ones(4,1);   %Stop the integration
direction = ones(4,1);    %Increasing directions,
                          %(as initially the error is zero)

%-------------------------------------------------------
function OP = outputs_fun( TH, X, TJ, XJ, Params )

n = Params.n;
A = Params.A;
P = Params.P;
c = Params.c;
Pbar = Params.Pbar;
Tsim = Params.Tf - Params.T0;

Xp = X(:,1:n);
Xh = X(:,n+1:2*n);
Xe = Xp - Xh;
T_jmp = X(:,2*n+6);
R_jmp = X(:,2*n+7);
Rtil_jmp = X(:,2*n+8);
Del_last = X(:,4*n+9);

q = 1;

Nevents = zeros(q,1);
Avg_int = zeros(q,1);
mITJ = -1*ones(q,1);

for i = 1:q
    Nevents(i) = length( TJ{i} );
    Avg_int(i) = Tsim/Nevents(i);
    if Nevents(i) >= 2
        mITJ(i) = min(TJ{i}(2:end) - TJ{i}(1:end-1));        
    end
end

disp('Tsim');
disp(Tsim);
disp('# updates');
disp( Nevents' );
disp('Avg interval.');
disp(Avg_int');
disp('min tx interval');
disp(mITJ');

%Set outputs to be returned
OP.n = n;
OP.V = sum(Xp'.*(P*Xp'))';            %Lyapunov function evolution
OP.Vd = X(:,2*n+1);
OP.hV = OP.V - OP.Vd;
OP.p_Lbnd = X(:,2*n+3);
OP.p = X(:,2*n+4);

OP.InfNXe = max(abs(Xe'))';

OP.de = Del_last;
for i = 1 : length(T_jmp)
    OP.de(i) = norm(expm(A*(TH(i,1)-T_jmp(i))),Inf).*Del_last(i);
end

OP.eps = OP.de ./ (c*sqrt(OP.Vd));
OP.Pbar = Pbar;

OP.Nevents = Nevents;
OP.Avg_int = Avg_int;
OP.mITJ = mITJ;

%-------------------------------------------------------
function OP = plot_fun( TH, X, TJ, XJ, OP, Par )

save_plot = Par.save_plot;

n = Par.n;
T0 = Par.T0;
Tf = Par.Tf;
Islot = Par.Islot;
Pbar = Par.Pbar;
R = Par.R;

%%
figure; h1 = gcf();
plot(TH(:,1),OP.V,'b',TH(:,1),OP.Vd,'b--','linewidth',2)
xlabel('$t$ (seconds)','interpreter','latex','fontsize',20);
legend('$V$', '$V_d$');
set(legend(), 'interpreter','latex','fontsize', 20)

figure; h2 = gcf();
ctmp = n*cumsum(XJ{1}(:,2*n+4));
plot(TJ{1},ctmp,'b-o','linewidth',2)
axis([0, 1.1*Tf, 0, 1.1*n*ctmp(end)])
ha0 = gca;
xlabel('$t$ (seconds)','interpreter','latex','fontsize',20);
ylabel('Total # bits transmitted','fontsize',20);
ha1 = axes('Position',[0.2 0.6 0.5 0.3])
plot(TJ{1}(6:8),ctmp(6:8),'b-o','linewidth',2)
set(ha1, 'Box', 'on')
%legend('$\bar{p} = 12$', '$\bar{p} = 20$');
%set(legend(), 'interpreter','latex','fontsize', 20)

figure; h3 = gcf();
stem(TJ{1},n*XJ{1}(:,2*n+4),'bo','linewidth',2)
hold on
stairs([T0, Islot],n*[Pbar, Pbar(end)],'b--','linewidth',2)
axis([0, 1.1*Tf, 0, n*max(Pbar)+1])
xlabel('$t$ (seconds)','interpreter','latex','fontsize',20);
ylabel('# bits transmitted','fontsize',20);

figure; h4 = gcf();
stairs([T0, Islot],n*[R, R(end)],'b--','linewidth',2)
%axis([0, 1.1*Tf, 0, 1.1*max(R)])
xlabel('$t$ (seconds)','interpreter','latex','fontsize',20);
ylabel('Channel comm. rate','fontsize',20);

1;
%%
%
if save_plot == 1
    figure(h1)
    saveas(gcf,'./fig_plots/V_Vd_conf','fig');
    set(gcf,'PaperPosition',[2.5 4 5.2 3.9]);
    print -depsc ./plots/V_Vd_conf
    close

    figure(h3)
    saveas(gcf,'./fig_plots/bit_profile_conf','fig');
    set(gcf,'PaperPosition',[2.5 4 5.2 3.9]);
    print -depsc ./plots/bit_profile_conf
    close
    
    figure(h4)
    saveas(gcf,'./fig_plots/R_conf','fig');
    set(gcf,'PaperPosition',[2.5 4 5.2 3.9]);
    print -depsc ./plots/R_conf
    close
    
    pause
    
    figure(h2)
    saveas(gcf,'./fig_plots/cum_bit_conf','fig');
    set(gcf,'PaperPosition',[2.5 4 5.2 3.9]);
    print -depsc ./plots/cum_bit_conf
    %close
end
%