%estimation of data capacity from the beginning of the j1 time slot to the
%end of j2 time slot, assuming that Tslot, R, Pbar satisfy J = 0

%Author: Pavan Tallapragada

function [ phiN, DataCap ] = data_cap_ind_J0( j1, j2, Par )

n = Par.n;
T0 = Par.T0;
Tslot = Par.Tslot;
Islot = Par.Islot;
Pbar = Par.Pbar;
R = Par.R;

Islot = [ T0, Islot ];

Tslot = Tslot(j1:j2);
Islot = Islot(j1:j2+1);
Pbar = Pbar(j1:j2);
R = R(j1:j2);

Tslot = Tslot(:);
Islot = Islot(:);
Pbar = Pbar(:);
R = R(:);

N = j2 - j1 + 1;
Nconstraints = N + N*(N-1)/2 + N;     %total number of constraints

c = - n*ones(N,1);                  %minus becasue linprog minimizes

A = zeros(Nconstraints, N);
b = zeros(Nconstraints, 1);

A1 = n*eye(N);
b1 = n*R.*Tslot + n*Pbar;
b1(Pbar==0) = 0;

A(1:N,:) = A1;
b(1:N) = b1;

A2 = n*R*(1./R)';
A2 = tril(A2);

b2 = n*R.*Islot(2:end) + n*Pbar;

ind = N;

for k = 1:N-1           %k is j1 in the paper
    nind = ind+N-k;
    A(ind+1:nind, k:end) = A2(k+1:end, k:end);
    b(ind+1:nind) = b2(k+1:end) - n*Islot(k)*R(k+1:end);
    ind = nind;
end

A3 = ones(N, 1)*(1./R)';
A3 = triu(A3);
b3 = Islot(end) - Islot(1:end-1);

A(ind+1:end,:) = A3;
b(ind+1:end,:) = b3;

phi_r = linprog(c,A,b,[],[],zeros(N,1),Inf*ones(N,1));

phiN = floor(phi_r);
DataCap = -c'*phiN;